# -*- coding: utf-8 -*-
from django.core import exceptions as django_exceptions
from django.db import models
from odin import exceptions as odin_exceptions
from odin.codecs import json_codec
import six
from baldr import form_fields


class ResourceField(six.with_metaclass(models.SubfieldBase, models.TextField)):
    """Field that serializes/de-serializes an Odin resource to the database seamlessly."""
    form_class = form_fields.ResourceField

    def __init__(self, resource_type, codec=json_codec, *args, **kwargs):
        super(ResourceField, self).__init__(*args, **kwargs)
        self.resource_type = resource_type
        self.codec = codec

    def to_python(self, value):
        if value in [None, '', {}, '{}']:  # Treat an empty JSON object as null.
            return

        if isinstance(value, self.resource_type):
            return value

        if isinstance(value, six.string_types):
            return self.codec.loads(value, self.resource_type, full_clean=False)

        raise django_exceptions.ValidationError(
            'Value provide is not a valid %s resource' % self.resource_type._meta.resource_name)

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value is None and not self.null:
            raise django_exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value in self.empty_values:
            raise django_exceptions.ValidationError(self.error_messages['blank'], code='blank')

        if isinstance(value, self.resource_type):
            try:
                value.full_clean()
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(ve.message_dict)

        raise django_exceptions.ValidationError(
            'Value provide is not a valid %s resource' % self.resource_type._meta.resource_name)

    def get_db_prep_save(self, value, connection):
        # Convert our JSON object to a string before we save
        if value is None:
            value = None if self.null else ""
        else:
            value = json_codec.dumps(value)
        return super(ResourceField, self).get_db_prep_save(value, connection=connection)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': self.form_class,
            'resource_type': self.resource_type
        }
        defaults.update(kwargs)
        return super(ResourceField, self).formfield(**defaults)


class ResourceListField(ResourceField):
    """Field that serializes/de-serializes a list of Odin resource to the database seamlessly."""
    form_class = form_fields.ResourceListField

    def to_python(self, value):
        if value in [None, '', {}, '{}']:  # Treat an empty JSON object as null.
            return

        if isinstance(value, (list, tuple)):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(str(ve.message_dict))
            except ValueError as ve:
                raise django_exceptions.ValidationError(ve.message)

        raise django_exceptions.ValidationError(
            'Value provide is not a valid %s resource' % self.resource_type._meta.resource_name)

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value is None and not self.null:
            raise django_exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value in self.empty_values:
            raise django_exceptions.ValidationError(self.error_messages['blank'], code='blank')

        if isinstance(value, (list, tuple)):
            errors = {}
            for idx, resource in enumerate(value):
                try:
                    super(ResourceField, self).validate(value, model_instance)
                except django_exceptions.ValidationError as ve:
                        errors[idx] = ve.message_dict
            if errors:
                raise django_exceptions.ValidationError(errors)

        else:
            raise django_exceptions.ValidationError(
                'Value provide is not a valid %s resource' % self.resource_type._meta.resource_name)


# Register field with south.
try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([
        (
            [ResourceField, ResourceListField],
            [],
            {
                'resource_type': ['resource_type', {}],
                'codec': ['codec', {'default': json_codec}],
            }
        )
    ], ["^baldr\.model_fields\.\w+Field"])
except ImportError:
    pass



# -*- coding: utf-8 -*-
from django.db import models
from odin import registration, Resource
from odin.mapping import FieldResolverBase


class ModelFieldResolver(FieldResolverBase):
    """
    Field resolver for Django Models
    """
    def get_field_dict(self):
        return {f.attname: f for f in self.obj._meta.fields}

registration.register_field_resolver(ModelFieldResolver, models.Model)



# -*- coding: utf-8 -*-
import odin


class Listing(odin.Resource):
    """
    Response for listing results. This includes offset, count support for paging etc.
    """
    class Meta:
        namespace = None

    # Wrapper to provide code completion
    def __init__(self, results, limit, offset=0, total_count=None):
        super(Listing, self).__init__(results, limit, offset, total_count)

    results = odin.ArrayField(help_text="The list of resources.")
    limit = odin.IntegerField(help_text="The resource limit in the result set.")
    offset = odin.IntegerField(default=0, help_text="The offset within the result set.")
    total_count = odin.IntegerField(null=True, help_text="The total number of items in the result set.")


class Error(odin.Resource):
    """
    Response returned for errors.
    """
    class Meta:
        namespace = None

    status = odin.IntegerField(help_text="HTTP status code of the response.")
    code = odin.IntegerField(help_text="An extended error code for more fine grained error results.")
    message = odin.StringField(help_text="A message that can be used for an end user.")
    developer_message = odin.StringField(null=True, help_text="More complex error message suitable for a developer.")
    meta = odin.StringField(null=True, help_text="Additional meta information that can help to solve issues.")



# -*- coding: utf-8 -*-
from django.core import exceptions as django_exceptions
from django.forms import widgets
from django.forms.fields import Field
from django.utils.translation import ugettext_lazy as _
import odin
from odin import exceptions as odin_exceptions
from odin.codecs import json_codec
import six


class ResourceField(Field):
    """
    Form field that wraps an Odin resource.
    """

    widget = widgets.Textarea

    default_error_messages = {
        'invalid': _('This field is not a valid %s resource.'),
    }

    def __init__(self, resource_type, codec=json_codec, indent=0, *args, **kwargs):
        assert issubclass(resource_type, odin.Resource)
        super(ResourceField, self).__init__(*args, **kwargs)

        self.resource_type = resource_type
        self.codec = codec
        self.indent = indent

    def prepare_value(self, value):
        if value is None:
            return ''
        elif isinstance(value, (list, tuple, odin.Resource)):
            return self.codec.dumps(value, indent=self.indent)
        else:
            return str(value)

    def to_python(self, value):
        if value is None or value == '':
            return

        if isinstance(value, self.resource_type):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(str(ve.message_dict))
            except ValueError as ve:
                raise django_exceptions.ValidationError(ve.message)

        raise django_exceptions.ValidationError(
            self.error_messages['invalid'] % self.resource_type._meta.resource_name, code='invalid')

    def validate(self, value):
        super(ResourceField, self).validate(value)

        if value is None:
            return

        if isinstance(value, self.resource_type):
            try:
                value.full_clean()
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(ve.message_dict)

        raise django_exceptions.ValidationError(
            self.error_messages['invalid'] % self.resource_type._meta.resource_name, code='invalid')


class ResourceListField(ResourceField):
    """
    Form field that wraps a list of odin fields.
    """
    def to_python(self, value):
        if value is None or value == '':
            return

        if isinstance(value, (list, tuple)):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(str(ve.message_dict))
            except ValueError as ve:
                raise django_exceptions.ValidationError(ve.message)

        raise django_exceptions.ValidationError(
            self.error_messages['invalid'] % self.resource_type._meta.resource_name, code='invalid')

    def validate(self, value):
        if value is None:
            return

        elif isinstance(value, (list, tuple)):
            errors = {}
            for idx, resource in enumerate(value):
                try:
                    super(ResourceField, self).validate(value)
                except django_exceptions.ValidationError as ve:
                        errors[idx] = ve.message_dict
            if errors:
                raise django_exceptions.ValidationError(errors)

        # Unknown type
        else:
            raise django_exceptions.ValidationError(
                self.error_messages['invalid'] % self.resource_type._meta.resource_name, code='invalid')



# coding=utf-8
from django.conf import settings
from django.conf.urls import url, patterns, include
from django.http import HttpResponse, Http404
from django.views.decorators.csrf import csrf_exempt
from odin.codecs import json_codec
from odin.exceptions import ValidationError
from baldr import content_type_resolvers
from baldr.exceptions import ImmediateErrorHttpResponse, ImmediateHttpResponse
from baldr.resources import Error, Listing


CODECS = {json_codec.CONTENT_TYPE: json_codec}
# Attempt to load other codecs that have dependencies
try:
    from odin.codecs import msgpack_codec
    CODECS[msgpack_codec.CONTENT_TYPE] = msgpack_codec
except ImportError:
    pass


class ResourceApiBase(object):
    # The resource this API is modelled on.
    resource = None

    # Handlers used to resolve a content-type from a request.
    # These are checked in the order defined until one returns a content-type
    content_type_resolvers = [
        content_type_resolvers.accepts_header(),
        content_type_resolvers.content_type_header(),
        content_type_resolvers.settings_default(),
    ]

    # Codecs that are supported for Encoding/Decoding resources.
    registered_codecs = CODECS

    def __init__(self, api_name=None):
        if api_name:
            self.api_name = api_name
        elif not hasattr(self, 'api_name'):
            self.api_name = "%ss" % self.resource._meta.name

    @property
    def urls(self):
        """
        Return url conf for resource object.
        """
        return patterns('', *self.base_urls())

    def resolve_content_type(self, request):
        """
        Resolve the request content type from the request.

        :returns: Identified content type; or ``None`` if content type is not identified.

        """
        for resolver in self.content_type_resolvers:
            content_type = resolver(request)
            if content_type:
                return content_type

    def handle_500(self, request, exception):
        """
        Handle *un-handled* exceptions

        :param request: The request object.
        :param exception: The exception that was un-handled.
        :return: An ``HttpError`` response resource.

        """
        # This is an unknown exception, return an unknown error message.
        if settings.DEBUG:
            # If we are in debug mode return more details and the stack-trace.
            import sys
            import traceback

            the_trace = '\n'.join(traceback.format_exception(*(sys.exc_info())))
            return Error(500, 50000, "An unknown error has occurred, the developers have been notified.",
                         str(exception), the_trace)
        else:
            return Error(500, 50000, "An unknown error has occurred, the developers have been notified.")

    def wrap_view(self, view):
        """
        This method provides the main entry point for URL mappings in the ``base_urls`` method.
        """
        @csrf_exempt
        def wrapper(request, *args, **kwargs):
            # Resolve content type used to encode/decode request/response content.
            content_type = self.resolve_content_type(request)
            try:
                request.codec = codec = self.registered_codecs[content_type]
            except KeyError:
                # This is just a plain HTTP response, we can't provide a rich response when the content type is unknown
                return HttpResponse(content="Content cannot be returned in the format requested.", status=406)

            callback = getattr(self, view)
            try:
                result = callback(request, *args, **kwargs)
            except Http404 as e:
                # Item is not found.
                status = 404
                resource = Error(status, 40400, str(e))
            except ImmediateHttpResponse as e:
                # An exception used to return a response immediately, skipping any further processing.
                response = HttpResponse(codec.dumps(e.resource), content_type=codec.CONTENT_TYPE, status=e.status)
                for key, value in (e.headers or {}).items():
                    response[key] = value
                return response
            except ValidationError as e:
                # Validation of a resource has failed.
                status = 400
                if hasattr(e, 'message_dict'):
                    resource = Error(status, 40000, "Fields failed validation.", meta=e.message_dict)
                else:
                    resource = Error(status, 40000, str(e))
            except NotImplementedError:
                # A mixin method has not been implemented, as defining a mixing is explicit this is considered a server
                # error that should be addressed.
                status = 501
                resource = Error(status, 50100, "This method has not been implemented.")
            except Exception as e:
                # Catch any other exceptions and pass them to the 500 handler for evaluation.
                resource = self.handle_500(request, e)
                status = resource.status
            else:
                if isinstance(result, tuple) and len(result) == 2:
                    resource, status = result
                else:
                    resource = result
                    status = 200
            return HttpResponse(codec.dumps(resource), content_type=codec.CONTENT_TYPE, status=status)
        return wrapper

    def base_urls(self):
        """
        Base URL mappings for this API.
        """
        return []

    def dispatch(self, request, request_type, **kwargs):
        """
        Primary method used to dispatch incoming requests to the appropriate method.
        """
        allowed_methods = getattr(self, "%s_allowed_methods" % request_type, [])
        request_method = self.method_check(request, allowed_methods)

        self.handle_authorisation(request, request_type, request_method)

        method = getattr(self, "%s_%s" % (request_method, request_type), None)
        if method is None:
            raise NotImplementedError()

        return method(request, **kwargs)

    def method_check(self, request, allowed):
        request_method = request.method.lower()

        if not allowed:
            raise Http404('`%s` not found.' % self.api_name)

        if request_method not in allowed:
            raise ImmediateErrorHttpResponse(405, 40500, "Method not allowed", headers={
                'Allow': ','.join(map(str.upper, allowed))
            })
        return request_method

    def handle_authorisation(self, request, request_type, request_method):
        """
        Evaluate if a request is authorised.

        :param request: The current Django request object.
        :param request_type: The current request type.
        :param request_method: The current request method.

        """
        pass


class ResourceApi(ResourceApiBase):
    """
    Provides an API that returns a specified resource object.
    """
    list_allowed_methods = ['get']
    detail_allowed_methods = ['get']

    def base_urls(self):
        return super(ResourceApi, self).base_urls() + [
            url(r'^%s/$' % self.api_name.lower(), self.wrap_view('dispatch_list')),
            url(r'^%s/(?P<resource_id>\d+)/$' % self.api_name.lower(), self.wrap_view('dispatch_detail')),
        ]

    def dispatch_list(self, request, **kwargs):
        return self.dispatch(request, 'list', **kwargs)

    def dispatch_detail(self, request, **kwargs):
        return self.dispatch(request, 'detail', **kwargs)


class ListMixin(ResourceApi):
    """
    Mixin to the resource API that provides a nice listing API.
    """
    def get_list(self, request):
        offset = int(request.GET.get('offset', 0))
        limit = int(request.GET.get('limit', 50))
        return Listing(
            self.list_resources(request, offset, limit),
            limit, offset
        )

    def list_resources(self, request, offset, limit):
        """
        Load resources
        :param limit: Resource count limit.
        :param offset: Offset within the list to return.
        :return: List of resource objects.
        """
        raise NotImplementedError


class CreateMixin(ResourceApi):
    """
    Mixin to the resource API to provide a Create API.
    """
    def __init__(self, *args, **kwargs):
        super(CreateMixin, self).__init__(*args, **kwargs)
        self.list_allowed_methods.append('post')

    def post_detail(self, request):
        resource = request.codec.loads(request.data, resource=self.resource)
        return self.create_resource(request, resource)

    def create_resource(self, request, resource):
        """
        Create method.
        """
        raise NotImplementedError


class RetrieveMixin(ResourceApi):
    """
    Mixin to the resource API to provide a Retrieve API.
    """
    def get_detail(self, request, resource_id):
        return self.retrieve_resource(request, resource_id)

    def retrieve_resource(self, request, resource_id):
        raise NotImplementedError


class UpdateMixin(ResourceApi):
    """
    Mixin to the resource API to provide a Update API.
    """
    def __init__(self, *args, **kwargs):
        super(UpdateMixin, self).__init__(*args, **kwargs)
        self.detail_allowed_methods.append('post')

    def post_detail(self, request, resource_id):
        resource = request.codec.loads(request.data, resource=self.resource)
        return self.update_resource(request, resource_id, resource)

    def update_resource(self, request, resource_id, resource):
        raise NotImplementedError


class DeleteMixin(ResourceApi):
    """
    Mixin to the resource API to provide a Delete API.
    """
    def __init__(self, *args, **kwargs):
        super(DeleteMixin, self).__init__(*args, **kwargs)
        self.detail_allowed_methods.append('delete')

    def post_detail(self, request, resource_id):
        return self.delete_resource(request, resource_id)

    def delete_resource(self, request, resource_id):
        raise NotImplementedError


class ApiCollection(object):
    """
    Collection of several resource API's.

    Along with helper methods for building URL patterns.
    """
    def __init__(self, *resource_apis, **kwargs):
        self.api_name = kwargs.pop('api_name', 'api')
        self.resource_apis = resource_apis

    @property
    def urls(self):
        if not hasattr(self, '_urls'):
            urls = []
            for resource_api in self.resource_apis:
                urls.extend(resource_api.urls)
            self._urls = urls
        return self._urls

    def include(self, namespace=None):
        return include(self.urls, namespace)

    def patterns(self):
        return [url(r'^%s/' % self.api_name, self.include())]
